"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createNameSpaceHandler = void 0;
const Debug = require("debug");
const genkey_service_1 = require("../services/genkey.service");
const cache_service_1 = require("../services/cache.service");
const pool_service_1 = require("../services/pool.service");
const Config_Context_1 = require("../services/Config.Context");
const utils_1 = require("../utils");
const proxy_1 = require("../proxy");
const osham_headers_1 = require("../osham.headers");
const pathToRegExp = require('path-to-regexp');
function createNameSpaceHandler(namespace, options) {
    const logger = Debug(`acp:handler(${namespace})`);
    logger(`Register ${namespace}`);
    const namespacePath = pathToRegExp(options.expose, [], {
        sensitive: false,
        strict: false,
        end: false,
    });
    const configContext = new Config_Context_1.ConfigContext(options.cache, options.rules);
    const proxyRequest = proxy_1.createProxy(options);
    return async function handler(ctx, next) {
        if (!namespacePath.test(ctx.path))
            return next();
        logger(`${ctx.path} matched!!`);
        const pathToCall = ctx.path.match(namespacePath)[1];
        logger(`${pathToCall} will be processed!`);
        const cacheConfig = configContext.getCacheConfig(pathToCall);
        const proxyPath = pathToCall + (ctx.search || '');
        if (ctx.method !== 'GET' || !cacheConfig) {
            logger(`${ctx.method} ${pathToCall} ${cacheConfig ? '(No cache config)' : ''}`);
            const proxyCtxN = await proxyRequest(proxyPath, ctx.method, ctx.headers);
            return proxyCtxN.pipes(ctx, osham_headers_1.OshamHeaders.notConfigured(ctx.method));
        }
        const cacheKey = genkey_service_1.generateKey(namespace, ctx, cacheConfig);
        const oshamHeaders = new osham_headers_1.OshamHeaders(cacheKey);
        logger(`Cache Check ${pathToCall}`);
        try {
            return await cache_service_1.Cache.get(cacheKey).then(utils_1.respondWithCtx(ctx, oshamHeaders.setHit(true).toRecords()));
        }
        catch (e) {
            oshamHeaders.setHit(false);
        }
        logger(`Cache miss ${pathToCall}`);
        if (!cacheConfig.pool) {
            logger(`No request pool`);
            const proxyCtxM = await proxyRequest(proxyPath, ctx.method, ctx.headers);
            proxyCtxM
                .toPromise()
                .then(res => cache_service_1.Cache.put(cacheKey, res.toJSON(), +cacheConfig.expires))
                .catch(() => ({}));
            return proxyCtxM.pipes(ctx, oshamHeaders.toRecords());
        }
        if (pool_service_1.RequestPool.has(cacheKey) && cache_service_1.Cache.isConnected()) {
            oshamHeaders.setPooled(true);
            return pool_service_1.RequestPool.wait(cacheKey).then(utils_1.respondWithCtx(ctx, oshamHeaders.toRecords()));
        }
        if (cache_service_1.Cache.isConnected()) {
            pool_service_1.RequestPool.add(cacheKey);
        }
        const proxyCtx = await proxyRequest(proxyPath, ctx.method, ctx.headers);
        proxyCtx
            .toPromise()
            .then(res => pool_service_1.RequestPool.putAndPublish(cacheKey, res.toJSON(), +cacheConfig.expires), error => {
            const response = utils_1.errorToData(error);
            pool_service_1.RequestPool.errorAndPublish(cacheKey, response);
        })
            .catch(error => {
            const response = utils_1.errorToData(error);
            pool_service_1.RequestPool.errorAndPublish(cacheKey, response);
        });
        return proxyCtx.pipes(ctx, oshamHeaders.setPooledMain(true).toRecords());
    };
}
exports.createNameSpaceHandler = createNameSpaceHandler;
