"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createProxy = void 0;
const http_1 = require("http");
const FollowRedirects = require("follow-redirects");
const Http = require("http");
const Https = require("https");
const url_1 = require("url");
const path_1 = require("path");
const Debug = require("debug");
const utils_1 = require("./utils");
const zlib_1 = require("zlib");
const NativeAgents = { http: Http, https: Https };
const logger = Debug('acp:service:proxy');
class Proxy {
    constructor(options) {
        this.options = options;
        this.isSecure = false;
        this.isSecure = Proxy.isSSL.test(options.target);
        this.target = url_1.parse(this.options.target);
        this.port = this.isSecure ? 443 : 80;
        const agents = this.options.followRedirects ? FollowRedirects : NativeAgents;
        this.agent = this.isSecure ? agents.https : agents.http;
    }
    request(path, method, headers, dataStream) {
        logger(`Request(${method}) ${path}`);
        const proxyCtx = {
            pipes,
            toPromise,
            proxy: this,
        };
        try {
            const options = this.getRequestOptions(path, method, headers);
            const request = this.agent.request(options);
            if (this.options.timeout) {
                request.setTimeout(this.options.timeout, () => request.abort());
            }
            if (dataStream && dataStream.pipe) {
                dataStream.pipe(request);
            }
            request.end();
            return new Promise(resolve => {
                request.on('response', (response) => {
                    resolve({ ...proxyCtx, request, response });
                });
                request.on('error', (e) => resolve({
                    ...proxyCtx,
                    request,
                    response: { ...Proxy.ErrorResponse },
                    message: e.message,
                }));
                request.on('abort', (e) => resolve({
                    ...proxyCtx,
                    request,
                    response: { ...Proxy.ErrorResponse },
                    message: e.message,
                }));
            });
        }
        catch (e) {
            return Promise.resolve({
                ...proxyCtx,
                message: e.message,
                response: { ...Proxy.ErrorResponse },
            });
        }
    }
    getRequestOptions(path, method, headers) {
        const options = {
            port: this.target.port || this.port,
            method,
            headers: { ...headers },
            host: this.target.hostname,
            path: path_1.join(this.target.path, path),
        };
        if (typeof options.headers.connection !== 'string' || !Proxy.upgradeHeader.test(options.headers.connection)) {
            options.headers.connection = 'close';
        }
        if (this.options.changeOrigin) {
            options.headers.host = options.headers.hostname = this.target.host;
        }
        if ((method === 'DELETE' || method === 'OPTIONS') && !headers['content-length']) {
            headers['content-length'] = '0';
            delete headers['transfer-encoding'];
        }
        return options;
    }
}
Proxy.isSSL = /^https/;
Proxy.upgradeHeader = /(^|,)\s*upgrade\s*($|,)/i;
Proxy.ErrorResponse = { statusCode: 503, statusMessage: 'Service Unavailable', headers: {} };
function createProxy(options) {
    const proxy = new Proxy(options);
    return (path, method, headers, dataStream) => proxy.request(path, method, headers, dataStream);
}
exports.createProxy = createProxy;
function pipes(ctx, osham_headers = {}) {
    const { response, message } = this;
    const { statusCode, statusMessage, headers } = response;
    ctx.statusCode = statusCode;
    ctx.statusMessage = statusMessage;
    ctx.responseHeaders = utils_1.writeHeaders(headers, ctx);
    utils_1.writeHeaders(osham_headers, ctx);
    ctx.body = message || response;
    return this;
}
function toPromise() {
    const { response, message } = this;
    const { statusCode, statusMessage } = response;
    const headers = utils_1.writeHeaders(response.headers);
    const ob = {
        data: '',
        statusCode,
        statusMessage,
        headers,
        ...this,
        toJSON() {
            return {
                data: this.data,
                statusCode: this.statusCode,
                statusMessage: this.statusMessage,
                headers: this.headers,
            };
        },
        message: '',
    };
    if (!(response instanceof http_1.IncomingMessage)) {
        return Promise.reject({ ...ob, message });
    }
    return new Promise((resolve, reject) => {
        const stream = withDecoder(ob);
        stream
            .on('error', onError)
            .on('end', onEnd)
            .on('data', (chunk) => (ob.data += chunk.toString()));
        return;
        function onError(e) {
            return reject({ ...ob, message: e.message });
        }
        function onEnd() {
            if (ob.isEncoded) {
                delete ob.headers['content-encoding'];
                if (ob.data) {
                    ob.headers['content-length'] = ob.data.length.toString();
                }
            }
            if (ob.statusCode >= 200 && ob.statusCode < 300) {
                return resolve(ob);
            }
            return reject({ ...ob, message });
        }
    });
}
function withDecoder(ob) {
    const encoding = ob.headers['content-encoding']?.trim();
    if (!encoding) {
        return ob.response;
    }
    ob.isEncoded = true;
    if (encoding === 'gzip') {
        return ob.response.pipe(zlib_1.createGunzip());
    }
    if (encoding === 'br') {
        return ob.response.pipe(zlib_1.createBrotliDecompress());
    }
    logger(`Encoding ${encoding} not handled properly!`);
    return ob.response;
}
